import { AxiosProxyConfig, AxiosResponse } from 'axios';
/**
 * Determines whether an object is a DavenportError.
 */
export declare function isDavenportError(error: any): error is DavenportError;
/**
 * A generic view document for listing and counting all objects in the database.
 */
export declare const GENERIC_LIST_VIEW: {
    "name": string;
    "map": string;
    "reduce": string;
};
/**
 * Configures a Davenport client and database by validating the CouchDB version, creating indexes and design documents, and then returning a client to interact with the database.
 */
export declare function configureDatabase<DocType extends CouchDoc>(databaseUrl: string, configuration: DatabaseConfiguration<DocType>, options?: ClientOptions): Promise<Client<DocType>>;
/**
 * A client for interacting with a CouchDB instance. Use this when you don't want or need to use the `configureClient` function to create a database and set up design docs or indexes.
 */
export declare class Client<T extends CouchDoc> {
    private baseUrl;
    private databaseName;
    private options;
    constructor(baseUrl: string, databaseName: string, options?: ClientOptions);
    private axios;
    private databaseUrl;
    private readonly jsonContentTypeHeaders;
    private getOption(name);
    /**
     * Checks that the Axios response is okay. If not, a DavenPort error is thrown.
     */
    private checkErrorAndGetBody(result);
    /**
     * Find matching documents according to the selector.
     */
    find(options: FindOptions<T>): Promise<T[]>;
    /**
     * Lists documents in the database. Warning: this result WILL list design documents, and it will force the `include_docs` option to false. If you need to include docs, use .listWithDocs.
     */
    listWithoutDocs(options?: ListOptions): Promise<ListResponse<{
        rev: string;
    }>>;
    /**
     * Lists documents in the database. Warning: this result WILL list design documents, and it will force the `include_docs` option to true. If you don't need to include docs, use .listWithoutDocs.
     */
    listWithDocs(options?: ListOptions): Promise<ListResponse<T>>;
    /**
     * Counts all documents in the database. Warning: this result WILL include design documents.
     */
    count(): Promise<number>;
    /**
     * Counts all documents by the given selector. Warning: this uses more memory than a regular count, because it needs to pull in the _id field of all selected documents. For large queries, it's better to create a dedicated view and use the .view function.
     */
    countBySelector(selector: DocSelector<T>): Promise<number>;
    countBySelector(selector: Partial<T>): Promise<number>;
    /**
     * Gets a document with the given id and optional revision id.
     */
    get(id: string, rev?: string): Promise<T>;
    /**
     * Inserts, updates or deletes multiple documents at the same time.
     *
     * Omitting the `_id` property from a document will cause CouchDB to generate the id itself.
     *
     * When updating a document, the `_rev` property is required.
     *
     * To delete a document, set the `_deleted` property to `true`.
     *
     * Note that CouchDB will return in the response an id and revision for every document passed as content to a bulk insert, even for those that were just deleted.
     *
     * If the `_rev` does not match the current version of the document, then that particular document will not be saved and will be reported as a conflict, but this does not prevent other documents in the batch from being saved.
     *
     * If the `newEdits` arg is `false` (to push existing revisions instead of creating new ones) the response will not include entries for any of the successful revisions (since their rev IDs are already known to the sender), only for the ones that had errors. Also, the `"conflict"` error will never appear, since in this mode conflicts are allowed.
     *
     * @param docs An array of documents that will be inserted, updated or deleted.
     * @param newEdits A boolean that determines whether to allow new edits or not.
     */
    bulk(docs: T[], newEdits?: boolean): Promise<BulkResponse>;
    /**
     * Creates a document with a random id. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    post(data: T): Promise<PostPutCopyResponse>;
    /**
     * Updates or creates a document with the given id. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    put(id: string, data: T, rev: string): Promise<PostPutCopyResponse>;
    /**
     * Copies the document with the given id and assigns the new id to the copy. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    copy(id: string, newId: string): Promise<PostPutCopyResponse>;
    /**
     * Deletes the document with the given id and revision id.
     */
    delete(id: string, rev: string): Promise<void>;
    /**
     * Checks that a document with the given id exists.
     */
    exists(id: string): Promise<boolean>;
    /**
     * Checks that a document that matches the field value exists.
     */
    existsByFieldValue(value: any, field: keyof T): Promise<boolean>;
    /**
     * Checks that a document matching the selector exists.
     */
    existsBySelector(selector: DocSelector<T>): Promise<boolean>;
    /**
     * Executes a view with the given designDocName and viewName. Will not reduce by default, pass in the { reduce: true } option to reduce.
     */
    view<DocType>(designDocName: string, viewName: string, options?: ViewOptions): Promise<ViewResult<DocType>>;
    /**
     * Executes a view with the given designDocName and viewName. This method will never reduce the result.
     */
    viewWithDocs<DocType>(designDocName: string, viewName: string, options?: ViewOptions): Promise<ViewResultWithDocs<DocType>>;
    /**
     * Creates the database associated with this client.
     */
    createDb(url?: string): Promise<CreateDatabaseResponse>;
    /**
     * Deletes the database associated with this client.
     */
    deleteDb(url?: string): Promise<BasicCouchResponse>;
    /**
     * Returns database info for the given database.
     */
    getDbInfo(url?: string): Promise<DbInfo>;
    private encodeOptions(options);
}
export default Client;
export declare class DavenportError extends Error {
    fullResponse: AxiosResponse;
    constructor(message: any, fullResponse: AxiosResponse);
    readonly isDavenport: boolean;
    status: number;
    statusText: string;
    url: string;
    body: any;
}
export interface CouchDoc {
    /**
     * The object's database id.
     */
    _id?: string;
    /**
     * The object's database revision.
     */
    _rev?: string;
}
export interface PostPutCopyResponse {
    id: string;
    rev: string;
}
export interface BulkDocumentError {
    id: string;
    error: "conflict" | "forbidden" | "unauthorized";
    reason: string | "Document update conflict.";
}
export declare type BulkResponse = (PostPutCopyResponse | BulkDocumentError)[];
export interface ViewOptions extends ListOptions {
    reduce?: boolean;
    group?: boolean;
    group_level?: number;
}
export interface ViewResult<DocType> {
    offset?: number;
    total_rows?: number;
    rows: ViewRow<DocType>[];
}
export interface ViewRow<DocType> {
    id?: string;
    key?: any;
    value: DocType;
}
export interface ViewRowWithDoc<DocType> extends ViewRow<DocType> {
    doc: DocType;
}
export interface ViewResultWithDocs<DocType> {
    offset?: number;
    total_rows?: number;
    rows: ViewRowWithDoc<DocType>[];
}
export declare type Key = string | number | Object;
/**
 * Options for listing database results.
 */
export interface ListOptions {
    limit?: number;
    key?: Key;
    keys?: Key[];
    start_key?: Key | Key[];
    end_key?: Key | Key[];
    inclusive_end?: boolean;
    descending?: boolean;
    skip?: number;
}
export interface AllDocsListResult<T> {
    rows: {
        id: string;
        key: string;
        value: {
            rev: string;
        };
        doc: T;
    }[];
    offset: number;
    total_rows: number;
}
export interface DbSizes {
    file: number;
    external: number;
    active: number;
}
export interface DbOther {
    data_size?: number;
}
export interface DbInfo {
    db_name: string;
    update_seq: string;
    sizes: DbSizes;
    purge_seq: number;
    other?: DbOther;
    doc_del_count: number;
    doc_count: number;
    disk_size: number;
    disk_format_version: number;
    data_size: number;
    compact_running: boolean;
    instance_start_time: number;
}
export interface BasicCouchResponse {
    ok: boolean;
}
export interface CreateDatabaseResponse extends BasicCouchResponse {
    /**
     * Whether the database already existed when trying to create it. Determined by CouchDB returning a 412 Precondition Failed response.
     */
    alreadyExisted: boolean;
}
export interface ListResponse<T> {
    offset: number;
    total_rows: number;
    rows: T[];
}
export interface FindOptions<T> {
    fields?: string[];
    sort?: Object[];
    limit?: number;
    skip?: number;
    use_index?: Object;
    selector: Partial<T> | DocSelector<T>;
}
export interface CouchDBView {
    map: string;
    reduce?: string;
}
export interface DesignDoc extends CouchDoc {
    views: {
        [name: string]: CouchDBView;
    };
    language: "javascript";
}
export interface DesignDocConfiguration {
    name: string;
    views: ({
        name: string;
    } & CouchDBView)[];
}
export interface DatabaseConfiguration<T> {
    name: string;
    indexes?: (keyof T)[];
    designDocs?: DesignDocConfiguration[];
}
export interface ClientOptions {
    /**
     * Whether the Davenport client should log warnings.
     */
    warnings?: boolean;
    /**
     * Username used to make requests with basic auth.
     */
    username?: string;
    /**
     * Password used to make requests with basic auth.
     */
    password?: string;
    /**
     * Proxy configuration object.
     */
    proxy?: AxiosProxyConfig;
}
export interface PropSelector {
    /**
     * Property is equal to this value.
     */
    $eq?: any;
    /**
     * Property is not equal to this value.
     */
    $ne?: any;
    /**
     * Property is greater than this value.
     */
    $gt?: any;
    /**
     * Property is greater than or equal to this value.
     */
    $gte?: any;
    /**
     * Property is less than this value.
     */
    $lt?: any;
    /**
     * Property is lesser than or equal to this value.
     */
    $lte?: any;
}
export declare type DocSelector<T> = Partial<Record<keyof T, PropSelector>>;
