"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = require("axios");
const logspect_1 = require("logspect");
/**
 * Indicates whether the request was a success or not (between 200-300).
 */
function isOkay(response) {
    return response.status >= 200 && response.status < 300;
}
/**
 * Determines whether an object is a DavenportError.
 */
function isDavenportError(error) {
    return error.isDavenport;
}
exports.isDavenportError = isDavenportError;
/**
 * A generic view document for listing and counting all objects in the database.
 */
exports.GENERIC_LIST_VIEW = {
    "name": "all",
    "map": function (doc) { emit(doc._id, doc); }.toString(),
    "reduce": "_count"
};
/**
 * Configures an instance of Axios with auth and validation defaults.
 */
function getAxiosInstance(options) {
    let auth;
    if (options && (options.username || options.password)) {
        auth = {
            username: options.username,
            password: options.password,
        };
    }
    const instance = axios_1.default.create({
        // Like fetch, Axios should never throw an error if it receives a response
        validateStatus: (status) => true,
        auth: auth,
        proxy: options.proxy
    });
    return instance;
}
/**
 * Configures a Davenport client and database by validating the CouchDB version, creating indexes and design documents, and then returning a client to interact with the database.
 */
function configureDatabase(databaseUrl, configuration, options) {
    return __awaiter(this, void 0, void 0, function* () {
        const ax = getAxiosInstance(options);
        const dbInfo = yield ax.get(databaseUrl);
        if (!isOkay(dbInfo)) {
            throw new Error(`Failed to connect to CouchDB instance at ${databaseUrl}. ${dbInfo.status} ${dbInfo.statusText}`);
        }
        const infoBody = dbInfo.data;
        const version = parseInt(infoBody.version);
        if (version < 2) {
            logspect_1.default(`Warning: Davenport expects your CouchDB instance to be running CouchDB 2.0 or higher. Version detected: ${version}. Some database methods may not work.`);
        }
        const putResult = yield ax.put(`${databaseUrl}/${configuration.name}`, undefined);
        const preconditionFailed = 412; /* Precondition Failed - Database already exists. */
        if (putResult.status !== preconditionFailed && !isOkay(putResult)) {
            throw new DavenportError(`${putResult.status} ${putResult.statusText} ${putResult.data}`, putResult);
        }
        if (Array.isArray(configuration.indexes) && configuration.indexes.length > 0) {
            const data = {
                index: {
                    fields: configuration.indexes
                },
                name: `${configuration.name}-indexes`,
            };
            const result = yield ax.post(`${databaseUrl}/${configuration.name}/_index`, data, {
                headers: {
                    "Content-Type": "application/json"
                },
            });
            if (!isOkay(result)) {
                throw new DavenportError(`Error creating CouchDB indexes on database ${configuration.name}.`, result);
            }
        }
        if (Array.isArray(configuration.designDocs) && configuration.designDocs.length > 0) {
            yield Promise.all(configuration.designDocs.map((designDoc) => __awaiter(this, void 0, void 0, function* () {
                const url = `${databaseUrl}/${configuration.name}/_design/${designDoc.name}`;
                const getDoc = yield ax.get(url);
                const okay = isOkay(getDoc);
                let docFromDatabase;
                if (!isOkay && getDoc.status !== 404) {
                    logspect_1.default(`Davenport: Failed to retrieve design doc "${designDoc.name}". ${getDoc.status} ${getDoc.statusText}`, getDoc.data);
                    return;
                }
                if (!isOkay) {
                    docFromDatabase = {
                        _id: `_design/${designDoc.name}`,
                        language: "javascript",
                        views: {}
                    };
                }
                else {
                    docFromDatabase = getDoc.data;
                }
                const docViews = designDoc.views;
                let shouldUpdate = false;
                docViews.forEach(view => {
                    if (!docFromDatabase.views || !docFromDatabase.views[view.name] || docFromDatabase.views[view.name].map !== view.map || docFromDatabase.views[view.name].reduce !== view.reduce) {
                        docFromDatabase.views = Object.assign({}, docFromDatabase.views, {
                            [view.name]: {
                                map: view.map,
                                reduce: view.reduce,
                            }
                        });
                        shouldUpdate = true;
                    }
                });
                if (shouldUpdate) {
                    logspect_1.default(`Davenport: Creating or updating design doc "${designDoc.name}" for database "${configuration.name}".`);
                    const result = yield ax.put(url, docFromDatabase, {
                        headers: {
                            "Content-Type": "application/json",
                        }
                    });
                    if (!isOkay(result)) {
                        logspect_1.default(`Davenport: Could not create or update CouchDB design doc "${designDoc.name}" for database "${configuration.name}". ${result.status} ${result.statusText}`, result.data);
                    }
                }
                return Promise.resolve();
            })));
        }
        return new Client(databaseUrl, configuration.name, options);
    });
}
exports.configureDatabase = configureDatabase;
/**
 * A client for interacting with a CouchDB instance. Use this when you don't want or need to use the `configureClient` function to create a database and set up design docs or indexes.
 */
class Client {
    constructor(baseUrl, databaseName, options = { warnings: true }) {
        this.baseUrl = baseUrl;
        this.databaseName = databaseName;
        this.options = options;
        this.databaseUrl = `${baseUrl}/${databaseName}/`;
        this.axios = getAxiosInstance(options);
    }
    get jsonContentTypeHeaders() {
        return {
            "Content-Type": "application/json"
        };
    }
    getOption(name) {
        if (!this.options) {
            return undefined;
        }
        return this.options[name];
    }
    /**
     * Checks that the Axios response is okay. If not, a DavenPort error is thrown.
     */
    checkErrorAndGetBody(result) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!isOkay(result)) {
                const message = `Error with ${result.config.method} request for CouchDB database ${this.databaseName} at ${result.config.url}. ${result.status} ${result.statusText}`;
                throw new DavenportError(message, result);
            }
            return result.data;
        });
    }
    ;
    /**
     * Find matching documents according to the selector.
     */
    find(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.post(`${this.databaseUrl}/_find`, options, {
                headers: this.jsonContentTypeHeaders
            });
            const body = yield this.checkErrorAndGetBody(result);
            if (body.warning && !!this.getOption("warnings")) {
                logspect_1.default("Davenport warning: Davenport.find result contained warning:", body.warning);
            }
            return body.docs;
        });
    }
    /**
     * Lists documents in the database. Warning: this result WILL list design documents, and it will force the `include_docs` option to false. If you need to include docs, use .listWithDocs.
     */
    listWithoutDocs(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(`${this.databaseUrl}/_all_docs`, {
                params: Object.assign({}, this.encodeOptions(options), { include_docs: false })
            });
            const body = yield this.checkErrorAndGetBody(result);
            return {
                offset: body.offset,
                total_rows: body.total_rows,
                rows: body.rows.map(r => r.value)
            };
        });
    }
    /**
     * Lists documents in the database. Warning: this result WILL list design documents, and it will force the `include_docs` option to true. If you don't need to include docs, use .listWithoutDocs.
     */
    listWithDocs(options = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(`${this.databaseUrl}/_all_docs`, {
                params: Object.assign({}, this.encodeOptions(options), { include_docs: true })
            });
            const body = yield this.checkErrorAndGetBody(result);
            return {
                offset: body.offset,
                total_rows: body.total_rows,
                rows: body.rows.map(r => r.doc)
            };
        });
    }
    /**
     * Counts all documents in the database. Warning: this result WILL include design documents.
     */
    count() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(`${this.databaseUrl}/_all_docs`, {
                params: {
                    limit: 0,
                }
            });
            const body = yield this.checkErrorAndGetBody(result);
            return body.total_rows;
        });
    }
    countBySelector(selector) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.find({
                fields: ["_id"],
                selector,
            });
            return result.length;
        });
    }
    /**
     * Gets a document with the given id and optional revision id.
     */
    get(id, rev) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(this.databaseUrl + id, {
                params: { rev }
            });
            const body = yield this.checkErrorAndGetBody(result);
            return body;
        });
    }
    /**
     * Inserts, updates or deletes multiple documents at the same time.
     *
     * Omitting the `_id` property from a document will cause CouchDB to generate the id itself.
     *
     * When updating a document, the `_rev` property is required.
     *
     * To delete a document, set the `_deleted` property to `true`.
     *
     * Note that CouchDB will return in the response an id and revision for every document passed as content to a bulk insert, even for those that were just deleted.
     *
     * If the `_rev` does not match the current version of the document, then that particular document will not be saved and will be reported as a conflict, but this does not prevent other documents in the batch from being saved.
     *
     * If the `newEdits` arg is `false` (to push existing revisions instead of creating new ones) the response will not include entries for any of the successful revisions (since their rev IDs are already known to the sender), only for the ones that had errors. Also, the `"conflict"` error will never appear, since in this mode conflicts are allowed.
     *
     * @param docs An array of documents that will be inserted, updated or deleted.
     * @param newEdits A boolean that determines whether to allow new edits or not.
     */
    bulk(docs, newEdits = true) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.post(this.databaseUrl + "_bulk_docs", { docs }, {
                headers: this.jsonContentTypeHeaders,
                params: { new_edits: newEdits }
            });
            const body = yield this.checkErrorAndGetBody(result);
            return body;
        });
    }
    /**
     * Creates a document with a random id. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    post(data) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.post(this.databaseUrl, data, {
                headers: this.jsonContentTypeHeaders
            });
            const body = yield this.checkErrorAndGetBody(result);
            return {
                id: body.id,
                rev: body.rev,
            };
        });
    }
    /**
     * Updates or creates a document with the given id. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    put(id, data, rev) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!rev && !!this.getOption("warnings")) {
                logspect_1.default(`Davenport warning: no revision specified for Davenport.put function with id ${id}. This may cause a document conflict error.`);
            }
            const result = yield this.axios.put(this.databaseUrl + id, data, {
                headers: this.jsonContentTypeHeaders,
                params: { rev }
            });
            const body = yield this.checkErrorAndGetBody(result);
            return {
                id: body.id,
                rev: body.rev,
            };
        });
    }
    /**
     * Copies the document with the given id and assigns the new id to the copy. By CouchDB convention, this will only return the id and revision id of the new document, not the document itself.
     */
    copy(id, newId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.request({
                url: this.databaseUrl + id,
                method: "COPY",
                headers: {
                    Destination: newId
                },
            });
            const body = yield this.checkErrorAndGetBody(result);
            return {
                id: body.id,
                rev: body.rev,
            };
        });
    }
    /**
     * Deletes the document with the given id and revision id.
     */
    delete(id, rev) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!rev && !!this.getOption("warnings")) {
                logspect_1.default(`Davenport warning: no revision specified for Davenport.delete function with id ${id}. This may cause a document conflict error.`);
            }
            const result = yield this.axios.delete(this.databaseUrl + id, {
                params: { rev }
            });
            yield this.checkErrorAndGetBody(result);
        });
    }
    /**
     * Checks that a document with the given id exists.
     */
    exists(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.head(this.databaseUrl + id);
            return result.status === 200;
        });
    }
    /**
     * Checks that a document that matches the field value exists.
     */
    existsByFieldValue(value, field) {
        return __awaiter(this, void 0, void 0, function* () {
            const findResult = yield this.find({
                fields: ["_id"],
                limit: 1,
                selector: {
                    [field]: value
                }
            });
            return findResult.length > 0;
        });
    }
    /**
     * Checks that a document matching the selector exists.
     */
    existsBySelector(selector) {
        return __awaiter(this, void 0, void 0, function* () {
            const findResult = yield this.find({
                fields: ["_id"],
                limit: 1,
                selector: selector,
            });
            return findResult.length > 0;
        });
    }
    /**
     * Executes a view with the given designDocName and viewName. Will not reduce by default, pass in the { reduce: true } option to reduce.
     */
    view(designDocName, viewName, options = { reduce: false }) {
        return __awaiter(this, void 0, void 0, function* () {
            // Ensure reduce is set to false unless explicitly set by the caller.
            if (typeof (options.reduce) !== "boolean") {
                options.reduce = false;
            }
            const result = yield this.axios.get(`${this.databaseUrl}_design/${designDocName}/_view/${viewName}`, {
                params: this.encodeOptions(options),
            });
            const body = yield this.checkErrorAndGetBody(result);
            return body;
        });
    }
    /**
     * Executes a view with the given designDocName and viewName. This method will never reduce the result.
     */
    viewWithDocs(designDocName, viewName, options = { reduce: false }) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(`${this.databaseUrl}_design/${designDocName}/_view/${viewName}`, {
                params: Object.assign({}, this.encodeOptions(options), { reduce: false, include_docs: true })
            });
            const body = yield this.checkErrorAndGetBody(result);
            return body;
        });
    }
    /**
     * Creates the database associated with this client.
     */
    createDb(url = this.databaseUrl) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.put(url);
            if (result.status === 412) {
                return {
                    ok: true,
                    alreadyExisted: true
                };
            }
            const body = yield this.checkErrorAndGetBody(result);
            return Object.assign({}, body, { alreadyExisted: false });
        });
    }
    /**
     * Deletes the database associated with this client.
     */
    deleteDb(url = this.databaseUrl) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.delete(url);
            return yield this.checkErrorAndGetBody(result);
        });
    }
    /**
     * Returns database info for the given database.
     */
    getDbInfo(url = this.databaseUrl) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.axios.get(url);
            return yield this.checkErrorAndGetBody(result);
        });
    }
    encodeOptions(options) {
        let keys = Object.getOwnPropertyNames(options || {});
        return keys.reduce((requestOptions, key) => {
            switch (key) {
                case "keys":
                case "key":
                case "start_key":
                case "end_key":
                    requestOptions[key] = JSON.stringify(options[key]);
                    break;
                default:
                    requestOptions[key] = options[key];
                    break;
            }
            return requestOptions;
        }, {});
    }
}
exports.Client = Client;
exports.default = Client;
class DavenportError extends Error {
    constructor(message, fullResponse) {
        super(message);
        this.fullResponse = fullResponse;
        this.isDavenport = true;
        this.status = fullResponse.status;
        this.statusText = fullResponse.statusText;
        this.body = fullResponse.data;
        this.url = fullResponse.headers.host || fullResponse.headers.HOST;
    }
}
exports.DavenportError = DavenportError;
;
//# sourceMappingURL=index.js.map